/* cx5x86.c: A (nasty) hack to screw around with the Cyrix 5x86 CPU. */
/*
    Written 1996 by Edwin Lim.

    Copyright 1996 Edwin Lim.

    This software may be used and distributed according to the terms
    of the GNU Public License, incorporated herein by reference.

    The author may be reached at elim@dodo.eng.uci.edu, unless
    he cannot be reached, at which he should be lurking around in
    the comp.os.linux.* news groups and an appropriate call would
    alert him of his own existence.  If he still cannot be reached,
    then it is safe to consider him dead for all practical purposes.

    This module is meant for those who are foolhardy enough to mess
    with their Cyrix 5x86 CPU.  May the force be with you.

*/

static const char *version =
    "cx5x86.c:v1.0c 5/27/96 Edwin Lim (elim@dodo.eng.uci.edu)\n";


#include <linux/config.h>

#include <linux/module.h>
#include <linux/version.h>

#include <linux/types.h>
#include <linux/fs.h>
#include <linux/errno.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/tty.h>
#include <linux/interrupt.h>
#include <linux/string.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/bitops.h>

#include "cx5x86.h"

/* 
   asm routine thanks to Bruce Stough <sbs1@PO6.RV.unisys.com>
   asm routine to set L1 WB bit in CR0, can only be done if code
   executes in privilege 0 level -- hopefully modules run in that
   level...
   
   Uncomment the #define if you want it.
*/

/* #define L1_WB */

#ifdef L1_WB
asm("movl %cr0,%eax; orl $0x20000000,%eax; movl %eax,%cr0");
#endif


static void setcxbit(const char *reg_name, int bit, int set_to)
{
        unsigned char ormask = 0x00; 
        unsigned char andmask = 0xff;
        unsigned char n; 
        unsigned char dummy; 
        unsigned char reg = 0x00;
        int i;

        /* Figure out what is the address of the register */
        for (i = 0; cxreg[i].name != NULL; i++)
                if (strcmp(reg_name, cxreg[i].name) == 0)
                        reg = cxreg[i].addr;
        
        /* If no such register exist, print error message */
        if (reg == 0) {
                printk("cx5x86: no such register: %s\n", reg_name);
                return;
        }
                
        /* Use left-shift to aim the bit... */
        for (n = 0x01, i = 0; i < bit; i++, n = (n << 1));
                if (set_to == 1)
                        ormask = n;
                else if (set_to == 0)
                        andmask = ~n;
        
        /* Write to the register */
        outb_p(reg,ADDRESS_PORT);	dummy = inb_p(DATA_PORT);
        dummy = (dummy & andmask) | ormask;
        outb_p(reg,ADDRESS_PORT);	outb_p(dummy,DATA_PORT);

        if (verb == 1) {
                printk("reg:%x bit:%d set_to:%d ", reg, bit, set_to);
                printk("OR:%x AND:%x\n", ormask, andmask);
        }
}


static void setcxreg(const char *reg_name, char *mask)
{
	unsigned char reg;
	unsigned char dummy;
	unsigned char ormask = 0x00;
	unsigned char andmask = 0xff;
	unsigned char n;
	int i;

        /* Figure out what is the address of the register */
        for (i = 0; cxreg[i].name != NULL; i++)
                if (strcmp(reg_name, cxreg[i].name) == 0)
                        reg = cxreg[i].addr;
        
	/* Figure out the proper OR and AND masks */
	for(n = 0x01, i = (strlen(mask) - 1 > 7) ? 7 : strlen(mask) - 1 ; 
                i >= 0; i--, n = (n << 1) ) {
		if (mask[i] == '1')
			ormask = ormask | n;
		else if (mask[i] == '0')
			andmask = andmask & ~n;
	}

	/* Write to the register */
	outb_p(reg, ADDRESS_PORT);	dummy = inb_p(DATA_PORT);
	dummy = (dummy & andmask) | ormask;
	outb_p(reg, ADDRESS_PORT);	outb_p(dummy,DATA_PORT);
  
        if (verb == 1) {
                printk("reg:%x ", reg);
                printk("mask:%s ", mask);
                printk("OR:%x AND:%x\n", ormask, andmask);
        }
}

static void showcxreg(void)
{
        int i;

        for (i = 0; cxreg[i].name != NULL; i++) {
                outb_p(cxreg[i].addr, ADDRESS_PORT);
                printk("%s=%x ", cxreg[i].name, inb_p(DATA_PORT));
        }
        printk("\n");

}

static void cxdefaults(void)
{
	/* 
	   This function is for compatibility purposes and for defaults,
	   if you decide to use them.
	*/

	/*
	   The lines below set the registers.  I have simplified it even
	   more this time.

	   Two methods:
	   ------------
	   There are two ways of setting the registers.  The old way is
	   to call setcxreg directly, the new way is to set the variables
	   (will elaborate later).

	   Difference between the two:
	   ---------------------------
	   There is an important difference between the two methods.
	   The old way (call setcxreg directly) immediately executes the
	   changes; the new way will just change the module variables
	   and then execute the changes after it returns from this
	   function call.

	   Old method:
	   ----------
	   All that needs to be done is to call the setcxreg() subroutine
	   with two arguments:

		setcxreg("REGISTER_ID", BIT_STRING);

	   REGISTER_ID is one of the predefined cpu registers like
	   CCR1, PCR0, etc.

           !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	   !!! NOTE NOTE NOTE NOTE NOTE NOTE NOTE NOTE NOTE NOTE NOTE !!!
           !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

	   NOTE:  There has been a change since version 002a!  The
		  setcxreg() function call now expect a char * value
		  for the REGISTER_ID.	That means that you have to put
		  quotes around the REGISTER_ID to make it into a string.
		  That is all you need to do.  For example:

                  setcxreg("CCR4","xxxx1xxx");

	   BIT_STRING is a character strings that represents the bits you
	   want to set, for example, "xxxx1xx0" will set the 0th bit to 0
	   and 3rd bit to 1, and leave the rest unchanged (so we hope...).

           New method:
	   -----------
	   Change the (global) variables to the settings you want.  The
	   new settings will override the values passed from the kernel
	   as our 'command line' argument, unless you speficied no
	   default in the command line.

	   E.g.,

	   btb_en = 1;

           or,

           pcr0 = "xxxx1xxx";

           The settings below are the defaults that I used in
           version 002a.
	*/
	
	/* For example, to enable branch target buffer */
	btb_en = 1;                            /* new method #1 */
        /* pcr0 = "xxxx1xxx"; */               /* new method #2 */
        /* setcxreg("PCR0","xxxxxx1x"); */     /* old method */

        mem_byp=1;      /* memory bypass */
        lsser=0;        /* load/store ordering to weak  (faster) */
        fp_fast=1;      /* fast fpu exception reporting */

        show_reg=1;
}

extern int init_module() {

 	char cputype, rev, step;
	char dummy;
        int i;

	/* Set the struct to point to externally passed in values */
        cx5bit[0].set_to = &use_smi;
        cx5bit[1].set_to = &smac;
        cx5bit[2].set_to = &mmac;
        cx5bit[3].set_to = &use_wbak;
        cx5bit[4].set_to = &lock_nw;
        cx5bit[5].set_to = &susp_hlt;
        cx5bit[6].set_to = &wt1;
        cx5bit[7].set_to = &bwrt;
        cx5bit[8].set_to = &use_susp;
        cx5bit[9].set_to = &smi_lock;
        cx5bit[10].set_to = &nmi_en;
        cx5bit[11].set_to = &linbrst;
        cx5bit[12].set_to = &smm_mode;
        cx5bit[13].set_to = &mapen0;
        cx5bit[14].set_to = &mapen1;
        cx5bit[15].set_to = &mapen2;
        cx5bit[16].set_to = &mapen3;
        cx5bit[17].set_to = &iort0;
        cx5bit[18].set_to = &iort1;
        cx5bit[19].set_to = &iort2;
        cx5bit[20].set_to = &mem_byp;
        cx5bit[21].set_to = &dte_en;
        cx5bit[22].set_to = &fp_fast;
        cx5bit[23].set_to = &cpuiden;
        cx5bit[24].set_to = &clk0;
        cx5bit[25].set_to = &clk1;
        cx5bit[26].set_to = &clk2;
        cx5bit[27].set_to = &hlf_clk;
        cx5bit[28].set_to = &perf0;
        cx5bit[29].set_to = &perf1;
        cx5bit[30].set_to = &perf2;
        cx5bit[31].set_to = &rstk_en;
        cx5bit[32].set_to = &btb_en;
        cx5bit[33].set_to = &loop_en;
        cx5bit[34].set_to = &lsser;

        cx5reg[0].set_to = &pcr0;
        cx5reg[1].set_to = &ccr1;
        cx5reg[2].set_to = &ccr2;
        cx5reg[3].set_to = &ccr3;
        cx5reg[4].set_to = &smar;
        cx5reg[5].set_to = &smar0;
        cx5reg[6].set_to = &smar1;
        cx5reg[7].set_to = &smar2;
        cx5reg[8].set_to = &ccr4;
        cx5reg[9].set_to = &pmr;
        cx5reg[10].set_to = &dir0;
        cx5reg[11].set_to = &dir1;

	/*
	   We have to check to see that we are dealing with a Cyrix.
	   It will abort if cpu is not the Cyrix 5x86.
	   Of course the abortion part has not been tested.
	   There is a more elaborate (and fail-safe) way of
	   testing for a Cyrix CPU, but that is left as an exercise
	   for the reader...
	*/
	printk("cx5x86: ");
	outb_p(DIR0,ADDRESS_PORT);	cputype = inb_p(DATA_PORT);
	outb_p(DIR1,ADDRESS_PORT);	rev = (inb_p(DATA_PORT) >> 4) & 0x0f;
	outb_p(DIR1,ADDRESS_PORT);	step = inb_p(DATA_PORT) & 0x0f;
	if (cputype == 0x1a)
		printk("cx486DX ");
	else if (cputype == 0x1d)
		printk("cx486DX/2 ");
	else if ( (cputype & 0xf0) == 0x20 )
		printk("cx5x86 ");
	else if ( (cputype & 0xf0) == 0x30 )
		printk("cx6x86 ");
	else {	
		printk("Cannot identify cpu!\n");
 		return 1;
	}
	printk("Revision %u.%u\n", rev, step); 
	if ( (cputype & 0xf0) == 0x20 ) {
                cxreg = cx5reg;
                cxbit = cx5bit;
        } else {
		printk("cx5x86: Not cx5x86.  Aborting.\n");
		return 1;
	}

	/*
	   MAPEN (bits 0-3 of CCR3) are set to 0001b.  This is 
	   necessary to access some of the registers.  It will be
	   changed back to 0000b before module exits.
	*/
	setcxreg("CCR3", "0001xxxx");

	/* This will printout some initial values */
        if (verb == 1) {
                printk("cx5x86: Initial register settings:\n");
                showcxreg();
        }

	/* Call defaults */
	if (use_def == 1)
                cxdefaults();

        /* Check to see if any reg is set, if so, act on it */
        for (i = 0; cxreg[i].name; i++)
                if ( strcmp(*cxreg[i].set_to, "") != 0 )
                        setcxreg(cxreg[i].name, *cxreg[i].set_to);

        /* Check to see if any bit is set, if so, act on it */
        for (i = 0; cxbit[i].bitname; i++)
                if ( *cxbit[i].set_to == 0 || *cxbit[i].set_to == 1)
                        setcxbit(cxbit[i].regname, 
                        cxbit[i].bitnumber,
                        *cxbit[i].set_to);

	/* Print out register settings after the changes, if any. */
        if (verb == 1) {
                printk("cx5x86: New register settings:\n");
                showcxreg();
        }

        /* Print out register if requested */
        if (show_reg == 1) {
                printk("cx5x86: Register settings:\n");
                showcxreg();
        }

        /* Set MAPEN bits to 0000b to prevent accidental changes */
	setcxreg("CCR3", "0000xxxx");

	return 0;
}

extern void cleanup_module(void) {}

